(******************************************************************************)
(*                                                                            *)
(*                                    Menhir                                  *)
(*                                                                            *)
(*   Copyright Inria. All rights reserved. This file is distributed under     *)
(*   the terms of the GNU General Public License version 2, as described in   *)
(*   the file LICENSE.                                                        *)
(*                                                                            *)
(******************************************************************************)

(* This extremely simple analysis restricts a grammar to the set
   of nonterminals that are reachable, via productions, from the
   start nonterminals. *)

val trim: BasicSyntax.grammar -> BasicSyntax.grammar

