(******************************************************************************)
(*                                                                            *)
(*                                    Menhir                                  *)
(*                                                                            *)
(*   Copyright Inria. All rights reserved. This file is distributed under     *)
(*   the terms of the GNU General Public License version 2, as described in   *)
(*   the file LICENSE.                                                        *)
(*                                                                            *)
(******************************************************************************)

open BasicSyntax

(** [inline g] traverses the grammar [g] and inlines away the nonterminal
    symbols whose definitions are marked [%inline]. The result is a grammar
    where no symbols are marked [%inline]. *)
val inline: grammar -> grammar
