(******************************************************************************)
(*                                                                            *)
(*                                    Menhir                                  *)
(*                                                                            *)
(*   Copyright Inria. All rights reserved. This file is distributed under     *)
(*   the terms of the GNU General Public License version 2, as described in   *)
(*   the file LICENSE.                                                        *)
(*                                                                            *)
(******************************************************************************)

include Set.S with type elt = string

val map: (elt -> elt) -> t -> t

(**[print] prints a set of strings as a comma-separated list,
   without opening and closing delimiters. *)
val print: t -> string
